DEFAULT_VICTORY_REWARD = 5
DEFAULT_BOREDOM_REWARD = 0
DEFAULT_DEFEAT_REWARD = -2
DEFAULT_ANGUISH_REWARD = -10

DEFAULT_EPISODE_LENGTH = 1
DEFAULT_N_GENERATIONS = 150
DEFAULT_N_TUNE_SAMPLES = 5
DEFAULT_TRAIN_BATCH_SIZE = 512 # 4_000
DEFAULT_N_AGENTS = 6
DEFAULT_IDENTITY_IS_LINEAR = False
DEFAULT_POLICY_SNAPSHOT_PERIOD = 100
DEFAULT_LEARNING_RATE = 2e-06
DEFAULT_N_ROLLOUT_WORKERS = 2
DEFAULT_FLIP_INITIAL_WEIGHTS = False
DEFAULT_DOMINANCE_RELATIONSHIP_POLARITY_THRESHOLD = 0.55
DEFAULT_OBSERVATION_ACCURACY = 1

DEFAULT_N_VISITOR_POPULATIONS_PER_N_VISITOR_AGENTS = 10
DEFAULT_N_TRANSPLANTS_PER_VISITOR_POPULATION = 30

DEFAULT_FREEZE_VISITORS = True